
from __future__ import annotations
from typing import Union, TypeVar, NamedTuple, Sequence, Deque, Optional, List, Type, Dict, Any
from types import ModuleType
from collections import namedtuple
import copy
import inspect
import xml.etree.ElementTree as ET
from xml.dom.minidom import Document
from xml.dom import minidom
import importlib
from collections import deque
import numpy as np

from .datatype import data_type
from .instanceattrs import InstanceAttrs
from .strcast import strcast
from .serializerutils import SerializerUtils
from .userinputnode import UserInputNode

#Generic class type alias
CLASS = TypeVar('CLASS')

#Generic class inst type alias
INSTANCE = TypeVar('INSTANCE')


class Serialize(SerializerUtils):

	def _delete_class_vars(self) -> None:
		if hasattr(self, 'doc'):
			del self.__class__.doc
			del self.__class__.rootName
			del self.__class__.root

	def _set_class_vars(self) -> None:
		self.__class__.doc = Document()
		self.__class__.rootName = self.__class__.__name__
		self.__class__.root = self.doc.createElement(self.rootName)
		self.doc.appendChild(self.root)

	def data_type(
			self, 
			data: Any,
			) -> str:
		return data.__class__.__name__

	def is_abstract_data_type(
			self, 
			data: Any,
			) -> Union[None, bool]:
		data_type = self.data_type(data)
		pdts = set([
			'int', 'float', 'bool', 
			'str', 'float64', 'float128' 
			'NoneType', 'float32', 'int64', 
			'int32', 'complex64', 'complex128'
		])
		if data_type in pdts:
			return
		return True

	def pdt_str_cast(
			self, 
			data: Union[
				int, str, float,
				bool, np.float64, np.float32, 
				np.int64, np.int32, None,
				np.complex64, np.complex128,
				np.float128,
			]
			) -> str:
		return str(data)

	def set_xml_attribs(
			self, 
			parent: ET.Element, 
			**kwargs,
			) -> None:
		for attrib_name, attrib_value in kwargs.items():
			parent.setAttribute(attrib_name, attrib_value)
	
	def set_non_hash_container_tag(
			self, 
			index: int, 
			minidom_doc: Document, 
			parent: ET.Element,
			) -> ET.Element:
		return minidom_doc.createElement((
			parent.tagName
			+ '_e'
			+ str(index)
		))

	def set_ndarray_tag(
			self, 
			index: int, 
			minidom_doc: Document, 
			parent: ET.Element,
			) -> ET.Element:
		return minidom_doc.createElement((
			parent.tagName
			+ '_e'
			+ str(index)
		))

	def set_dict_tag(
			self, 
			key, 
			minidom_doc, 
			parent
		) -> ET.Element:
		tag = self.doc.createElement(key)
		return tag
			
	def serialize_pdt(
			self, 
			data, 
			minidom_doc, 
			parent,
		) -> None: 
		xml_str = self.pdt_str_cast(data)
		tag = minidom_doc.createTextNode(xml_str)
		parent.setAttribute('type', self.data_type(data))
		parent.appendChild(tag)

	def serialize_struct(
			self, 
			struct, 
			minidom_doc, 
			parent,
		) -> None:
		typ = data_type(struct)
		if typ == 'instance':
			self.serialize_inst(struct, minidom_doc, parent)
		elif typ == 'namedtuple':
			self.serialize_namedtuple(struct, minidom_doc, parent)
		elif typ == 'list':
			self.serialize_list(struct, minidom_doc, parent)
		elif typ == 'tuple':
			self.serialize_tuple(struct, minidom_doc, parent)
		elif typ == 'ndarray':
			self.serialize_ndarray(struct, minidom_doc, parent)
		elif typ == 'dict':
			self.serialize_dict(struct, minidom_doc, parent)

	def serialize_inst(
			self, 
			cls, 
			minidom_doc, 
			parent,
		) -> None:
		for attr_name, attr_value in self.get_instance_dict(cls).items():
			attr_value = self.get_instance_dict(cls)[attr_name]
			tag = minidom_doc.createElement(attr_name)
			self.set_xml_attribs(
				parent, 
				type = 'instance', 
				class_name = self.get_object_name(cls),
				module = self.get_module_name(cls)
			) 
			parent.appendChild(tag)
			self._serialize(attr_value, tag)

	def serialize_namedtuple(
			self, 
			namedtuple_: NamedTuple, 
			minidom_doc: Document, 
			parent: ET.Element,
		) -> None:
		for field_key, field_val in namedtuple_._asdict().items():
			if field_key.startswith('__'):
				continue
			tag = minidom_doc.createElement(field_key)
			self.set_xml_attribs(
				parent, 
				type = 'namedtuple',
				namedtuple_name = self.get_object_name(namedtuple_) 
			)
			parent.appendChild(tag)
			self._serialize(field_val, tag)

	def serialize_list(
			self, 
			list_: List, 
			minidom_doc: Document, 
			parent: ET.Element,
		) -> None:
		for i, e in enumerate(list_):
			tag = self.set_non_hash_container_tag(
				i, minidom_doc, parent
			) 
			parent.appendChild(tag)
			self.set_xml_attribs(parent, type = 'list')
			parent.setAttribute('type', 'list')
			self._serialize(e, tag)

	def serialize_tuple(
			self, 
			tuple_: Tuple, 
			minidom_doc: Document, 
			parent: ET.Element,
		) -> None:
		for i, e in enumerate(tuple_):
			tag = self.set_non_hash_container_tag(
				i, minidom_doc, parent
			) 
			parent.appendChild(tag)
			self.set_xml_attribs(parent, type = 'tuple')
			parent.setAttribute('type', 'tuple')
			self._serialize(e, tag)

	def serialize_ndarray(
			self, 
			ndarray_: np.ndarray, 
			minidom_doc: Document, 
			parent: ET.Element,
		) -> None:
		for i, e in enumerate(ndarray_):
			tag = self.set_non_hash_container_tag(
				i, minidom_doc, parent
			) 
			parent.appendChild(tag)
			self.set_xml_attribs(parent, type = 'ndarray')
			parent.setAttribute('type', 'ndarray')
			self._serialize(e, tag)

	def make_dict_keys_serializeable(
			self,
			dict_: Dict,
	) -> Dict:
		dtype_suffix_dict = {
			'int' : '',
			'str' : '',
			'float' : '',
			'int32' : '-INT32',
			'int64' : '-INT64',
			'float32' : '-FLOAT32',
			'float64' : '-FLOAT64',
			'float128' : '-FLOAT128',
			'complex64' : '-COMPLEX64',
			'complex128' : '-COMPLEX128',
		}
		serializeable_dict = {}
		for key, value in dict_.items():
			if isinstance(key, str):
				serializeable_dict[key] = value 
			elif isinstance(key, tuple):
				prefix = 'key_'
				suffix = 'tuple'
				serializeable_key = prefix 
				for i in key:
					serializeable_key += str(i) + dtype_suffix_dict[i.__class__.__name__] + '_'
				serializeable_key = serializeable_key + suffix 
				serializeable_dict[serializeable_key] = value 
			else:
				serializeable_key = (
					f'key_{key}_'
					+ f'{dtype_suffix_dict[key.__class__.__name__]}' 
				)
				serializeable_dict[serializeable_key] = value 
		return serializeable_dict
				
	def serialize_dict(
			self, 
			dict_: Dict, 
			minidom_doc: Document, 
			parent: ET.Element,
		) -> None:
		dict_ = self.make_dict_keys_serializeable(dict_)
		for key in dict_:
			tag = self.set_dict_tag(key, minidom_doc, parent)
			parent.appendChild(tag)
			self.set_xml_attribs(parent, type = 'dict')
			self._serialize(dict_[key], tag)

	def add_child(
			self, 
			data: Any, 
			minidom_doc: Document, 
			parent: ET.Element,
		) -> None:
		if self.data_type(data) == 'NoneType':
			return
		elif self.is_abstract_data_type(data):
			self.serialize_struct(data, minidom_doc, parent)
		else:
			self.serialize_pdt(data, minidom_doc, parent)
		

	def _serialize(
			self, 
			value: Any, 
			parent: ET.Element,
		) -> None:
		self.add_child(value, self.doc, parent)

	def serialize(
			self, 
			xml_filename: Optional[str] = None,
		) -> None:
		self._set_class_vars()
		parent = self.root
		self._serialize(self, parent)
		if xml_filename is not None:
			self._write_xml(xml_filename)
		self._delete_class_vars()

	def _write_xml(
			self, 
			xml_filename: str,
		) -> None:	  
		xmlstr = self.doc.toprettyxml(indent="	  ")
		xml_file = open(xml_filename, 'w')
		xml_file.write(xmlstr)
		xml_file.close()





from __future__ import annotations
from typing import Union, TypeVar, NamedTuple, Sequence, Deque, Optional, List, Type, Dict, Any
from types import ModuleType
import xml.etree.ElementTree as ET
from .strcast import StrCastUtils as StrCastUtils
from xml.dom.minidom import Document
from xml.dom import minidom
import importlib

from .datatype import data_type

#Generic class type alias
CLASS = TypeVar('CLASS')

#Generic class inst type alias
INSTANCE = TypeVar('INSTANCE')

class SerializerUtils(object):

	def get_object_name(
			self,
			object_: Union[INSTANCE, NamedTuple], 
			) -> str:
		return object_.__class__.__name__ 

	def get_module_name(
			self,
			class_: CLASS
			):
		return class_.__module__

	def get_instance_dict(
			self, 
			inst: INSTANCE,
			keys: Optional[bool] = False,
			values: Optional[bool] = False,
			) -> Union[Dict, List]:
		if keys:
			return list(inst.__dict__)
		elif values:
			return list(inst.__dict__.values())
		return inst.__dict__ 

	def get_namedtuple_dict(
			self, 
			namedtuple_: NamedTuple,
			):
		return namedtuple_._asdict() 
	
	def get_module(
			self, 
			xml_node: ET.Element,
			) -> ModuleType:
		if 'module' in xml_node.attrib:
			module_name = self.get_xml_node_attr(xml_node, 'module')
			return importlib.import_module(module_name)
		return importlib.import_module('__main__')
	
	def get_xml_node_attr(
			self, 
			xml_node: ET.Element,
			xml_node_attr_name: str,
			) -> str:
		return xml_node.attrib[xml_node_attr_name]

	def set_xml_node_attr(
			self, 
			xml_node: ET.Element,
			xml_node_attr_name: str,
			xml_node_attr_value: str,
			) -> None:
		xml_node.attrib[xml_node_attr_name] = xml_node_attr_value

	def null_adt(
			self,
			size: Optional[int] = None,
			keys: Optional[Union[int, str]] = None,
			key_attr: Optional[str] = None, 
			) -> Union[List, Dict]:
		if size is not None:
			return self.null_list(size)
		return self.null_dict(keys, key_attr)

	def null_list(
			self,
			size: Optional[int] = None,
			) -> List:
		return [None for element in range(size)]

	def recover_xml_dict_key(
			self,
			key: str,
	) -> Union[Tuple, int, float]:
		if 'key_' not in key:
			return key
		key = key.split('_')[1:]
		str_cast = StrCastUtils()
		if key[-1] == 'tuple':
			key.pop()
			key = tuple((str_cast.std_str_cast(i.replace('-', '_')) for i in key))
			return key
		return str_cast.std_str_cast(key[0].replace('-', '_'))

	def null_dict(
			self,
			keys: Optional[Any] = None,
			key_attr: Optional[str] = None, 
			) -> Dict:
		if key_attr is not None:
			keys = [
				self.recover_xml_dict_key(getattr(inst, key_attr)) 
				for inst in keys
			]
			return {key : None for key in keys}
		return {
			self.recover_xml_dict_key(key) : None 
			for key in keys
		}
		

	def class_instance(
			self, 
			xml_node: ET.Element,
			) -> INSTANCE:
		module = self.get_module(xml_node)
		class_name = None 
		try:
			class_name = self.get_xml_node_attr(xml_node, 'class_name')
		except:
			class_name = self.get_xml_node_attr(xml_node, 'class')
		class_inst = getattr(module, class_name).__new__(
			getattr(module, class_name)
		)
		return class_inst






















from __future__ import annotations
import numpy as np
from typing import Union, TypeVar, NamedTuple, Sequence, Deque, Optional, List, Type, Dict, Any, Type
from types import ModuleType
from collections import namedtuple
import copy
import inspect
import xml.etree.ElementTree as ET
from xml.dom.minidom import Document
from xml.dom import minidom
import importlib
from collections import deque

from .datatype import data_type
from .instanceattrs import InstanceAttrs
from .strcast import strcast
from .serializerutils import SerializerUtils
from .userinputnode import UserInputNode
from .serialize import Serialize


#Generic class type alias
CLASS = TypeVar('CLASS')

#Generic class inst type alias
INSTANCE = TypeVar('INSTANCE')

#Add documentation...otherwise good



class DeserializeUserInput(SerializerUtils):

	def init_args(
			self, 
			inst: INSTANCE,
		) -> None:
		class_ = type(inst)
		signature = inspect.signature(class_)
		init_arg_names = [
			arg_name for arg_name in list(signature.parameters)
		]
		inst_attr_names = self.get_inst_dict(keys=True) 
		

	def instance_attr(
			self, 
			node: UserInputNode, 
			xml_node: ET.Element, 
			inst_attrs: InstanceAttrs,
		) -> None:
		module = importlib.import_module('__main__')
		inst = self.class_instance(xml_node)
		new_inst_attrs = InstanceAttrs(inst)
		empty_str_attrs = []
		for attr in new_inst_attrs.default_instance_attrs():
			if new_inst_attrs.is_empty_str(attr):
				empty_str_attrs.append(attr)
		inst_attrs.call_init()
		setattr(self, node.name, inst)
		inst.__deserialize_user_input(
			node, 
			xml_node, 
			empty_str_attrs, 
			new_inst_attrs
		)

	def instance_list_attr(
			self, 
			node: UserInputNode, 
			xml_node: ET.Element, 
			inst_attrs: InstanceAttrs,
		) -> None:
		num_children = len(xml_node)
		null_list = self.null_adt(num_children)
		inst_list = []
		children = []
		inst_list_empty_str_attrs = []
		inst_list_attrs = []
		for index in range(len(null_list)):
			inst = self.class_instance(xml_node[index])
			new_inst_attrs = InstanceAttrs(inst)
			empty_str_attrs = []
			for attr in new_inst_attrs.default_instance_attrs():
				if new_inst_attrs.is_empty_str(attr):
					empty_str_attrs.append(attr)
			inst_list.append(inst)
			inst_list_attrs.append(new_inst_attrs)
			inst_list_empty_str_attrs.append(empty_str_attrs)
			children.append(node.add_child(xml_node[index]))
		inst_attrs.call_init()
		setattr(self, node.name, inst_list) 
		for index in range(len(children)):
			inst_list[index].__deserialize_user_input(
				children[index], 
				xml_node[index], 
				inst_list_empty_str_attrs[index], 
				inst_list_attrs[index]
			)

	def numpy_dtype_dict(
			self,
			) -> Dict[str, Union[
				Type[np.ndarray],
				Type[np.float32], 
				Type[np.float64], 
				Type[np.int32],
				Type[np.int64], 
				Type[np.complex64],
				Type[np.complex128]
			]]:
		return {
			"array" : np.array,
			"float32" : np.float32,
			"float64" : np.float64,
			"int32" : np.int32,
			"int64" : np.int64,
			"complex64" : np.complex64,
			"complex128" : np.complex128
		}

	def ndarray_dtype(
			self,
			array_dtype: str,
			) -> Union[
				np.float32, 
				np.float64, 
				np.int32,
				np.int64, 
				np.complex64,
				np.complex128
			]:
		dtype_name = array_dtype.replace(" array", "")
		if dtype_name == "array":
			return self.numpy_dtype_dict()["float32"]
		return self.numpy_dtype_dict()[dtype_name]
			

	def cast_to_numpy_dtype(
			self,
			val: Union[List, int, float],
			dtype: str,
			) -> Union[
				np.ndarray,
				np.float32, 
				np.float64, 
				np.int32,
				np.int64, 
				np.complex64,
				np.complex128
			]:
		if "array" in dtype:
			array_dtype = self.ndarray_dtype(dtype)
			return self.numpy_dtype_dict()['array'](
				val, 
				dtype=array_dtype
			)
		return self.numpy_dtype_dict()[dtype](
			val 
		)
		

	def non_instance_attr(
			self, 
			node: UserInputNode, 
			xml_node: ET.Element, 
			empty_str_attrs: List, 
			inst_attrs: InstanceAttrs,
		) -> None:
		if (type(xml_node.text) == type(None)
				and xml_node.tag in empty_str_attrs):
			setattr(self, node.name, '')
			inst_attrs.call_init()
		else:
			if "numpy" in xml_node.attrib:
				setattr(
					self, 
					node.name, 
					self.cast_to_numpy_dtype(
						strcast(xml_node.text),
						xml_node.attrib["numpy"]
					)
				)
			else:
				setattr(
					self, 
					node.name, 
					strcast(xml_node.text)
				)
			inst_attrs.call_init()

	def __deserialize_user_input(
			self, 
			node: UserInputNode, 
			xml_node: ET.Element, 
			empty_str_attrs: List, 
			inst_attrs: InstanceAttrs,
			) -> UserInputNode:
		for xml_child in node.xml_node:
			child = node.add_child(xml_child)
			#Added "or child.type == 'class'" for legacy support
			if child.type_ == 'instance' or child.type_ == 'class': 
				self.instance_attr(child, xml_child, inst_attrs)
			elif child.type_ == 'instance_list':
				self.instance_list_attr(child, xml_child, inst_attrs)
			elif child.type_ == 'non_instance':
				self.non_instance_attr(
					child, 
					xml_child, 
					empty_str_attrs, 
					inst_attrs,
				)
		return node
			

	def deserialize_user_input(
			self, 
			node: UserInputNode, 
			xml_root: ET.Element, 
			empty_str_attrs: List, 
			inst_attrs: InstanceAttrs,
		) -> str:
		name = xml_root.tag
		root = UserInputNode(xml_root, name, 'instance')
		self.__deserialize_user_input(
			root, 
			xml_root, 
			empty_str_attrs, 
			inst_attrs, 
		)
		return root
'''
-----------------------------------------------------------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------------------------------
'''



from __future__ import annotations
from typing import Union, \
	TypeVar, \
	NamedTuple, \
	Sequence, \
	Deque, \
	Optional, \
	List, \
	Type, \
	Dict, \
	Any, \
	Tuple, \
	Generator, \
	Set
from types import ModuleType
from collections import namedtuple
import gc
import copy
import inspect
import sys
import xml.etree.ElementTree as ET
from xml.dom.minidom import Document
from xml.dom import minidom
import importlib
from collections import deque
import numpy as np
from numpy import ndarray

from .datatype import data_type
from .instanceattrs import InstanceAttrs
from .strcast import strcast
from .serializerutils import SerializerUtils
from .userinputnode import UserInputNode
from .serialize import Serialize
from .deserializeuserinput import DeserializeUserInput

#Generic class type alias
CLASS = TypeVar('CLASS')

#Generic class inst type alias
INSTANCE = TypeVar('INSTANCE')

class Deserialize(DeserializeUserInput):

	def set_path(
			self, 
			path: List, 
			index: int,
		) -> List:
		child_path = copy.deepcopy(path)
		child_path.append(index)
		return child_path

	def tag_namedtuple(
			self, 
			node,
			):
		node.tag = (
			node.tag 
			+ '___' 
			+ self.get_xml_node_attr(node, 'namedtuple_name')
		)
		self.set_xml_node_attr(node, 'type', 'dict')
		return node

	def tag_child_namedtuples(
			self, 
			node,
			):
		for child in node:
			if self.is_namedtuple(child):
				self.tag_namedtuple(child)

	def is_namedtuple(
			self, 
			node,
			):
		if self.attr_type(node) == 'namedtuple':
			return True

	def deserialize_nested_attr(
			self, 
			name, 
			val, 
			path,
			):
		if isinstance(path, list):
			path = deque(path)
		return self.deserialize_nested_attr_(
			self.get_instance_dict(self)[name], 
			val, 
			path
		)

	def deserialize_nested_attr_(
			self, 
			attr, 
			val, 
			path,
			):
		if len(path) > 1:
			pos = path.popleft()
			return self.deserialize_nested_attr_(
				attr[pos], 
				val, 
				path
			)
		attr[path[0]] = val
		return attr[path[0]] 
	   
	def deserialize_list(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if len(path) != 0:
			null_list = [None for gchild in child]
			name = node.tag
			self.deserialize_nested_attr(name, null_list, path)
			for index, grandchild in enumerate(child):
				grandchild_path = self.set_path(path, index) 
				self.deserialize_child(node, grandchild, grandchild_path)
		else:
			null_list = self.null_adt(size=len(node)) 
			name = node.tag
			setattr(self, name, null_list)
			for index, child in enumerate(node):
				child_path = self.set_path(path, index) 
				self.deserialize_child(node, child, child_path)

	def list_to_tuple(
			self,
			val: Any
			) -> Tuple:
		return (
			tuple(map(self.list_to_tuple, val)) 
				if type(val) == tuple or type(val) == list 
				else val 
		)

	def list_to_ndarray(
			self,
			val: Any
			) -> Tuple:
		return (
			np.array(list(map(self.list_to_tuple, val))) 
				if type(val) == tuple or type(val) == list 
				else val 
		)
	
	def immutable_is_deserialized(
			self,
			tmp_tuple_attr_list: Union[List, np.ndarray],
			) -> bool:
		if isinstance(tmp_tuple_attr_list, ndarray):
			tmp_tuple_attr_list = tmp_tuple_attr_list.tolist()
		if None in list(self.flatten_container(
			tmp_tuple_attr_list
		)):
			return False
		return True

	def flatten_container(
		self,
		container: Union[List, Tuple, Set],
		) -> Generator[List, Tuple, Set]:
		for element in container:
			if isinstance(element, (list,tuple)):
				for nested_element in self.flatten_container(
					element
				):
					yield nested_element
			else:
				yield element

	def deserialize_tuple(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if len(path) != 0:
			null_list = [None for gchild in child]
			name = node.tag
			self.deserialize_nested_attr(name, null_list, path)
			for index, grandchild in enumerate(child):
				grandchild_path = self.set_path(path, index) 
				self.deserialize_child(node, grandchild, grandchild_path)
			tmp_tuple_attr_list = getattr(self, name)
		else:
			null_list = self.null_adt(size=len(node)) 
			name = node.tag
			setattr(self, name, null_list)
			for index, child in enumerate(node):
				child_path = self.set_path(path, index) 
				self.deserialize_child(node, child, child_path)
		tmp_tuple_attr_list = getattr(self, name)
		if self.immutable_is_deserialized(tmp_tuple_attr_list):
			setattr(self, name, self.list_to_tuple(tmp_tuple_attr_list))

	def deserialize_ndarray(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if len(path) != 0:
			null_list = [None for gchild in child]
			name = node.tag
			self.deserialize_nested_attr(name, null_list, path)
			for index, grandchild in enumerate(child):
				grandchild_path = self.set_path(path, index) 
				self.deserialize_child(node, grandchild, grandchild_path)
			tmp_ndarray_attr_list = getattr(self, name)
		else:
			null_list = self.null_adt(size=len(node)) 
			name = node.tag
			setattr(self, name, null_list)
			for index, child in enumerate(node):
				child_path = self.set_path(path, index) 
				self.deserialize_child(node, child, child_path)
		tmp_ndarray_attr_list = getattr(self, name)
		if self.immutable_is_deserialized(tmp_ndarray_attr_list):
			setattr(self, name, self.list_to_ndarray(tmp_ndarray_attr_list))


	def deserialize_dict(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if len(path) != 0:
			name = node.tag
			null_dict = self.null_adt(
				keys=child, 
				key_attr='tag'
			)  
			self.deserialize_nested_attr(name, null_dict, path)
			for grandchild in child:
				grandchild.tag = self.recover_xml_dict_key(grandchild.tag)
				grandchild_path = self.set_path(path, grandchild.tag) 
				self.deserialize_child(node, grandchild, grandchild_path)
		else:
			null_dict = self.null_adt(
				keys=node, 
				key_attr='tag'
			)  
			name = node.tag
			setattr(self, name, null_dict)
			for child in node:
				child.tag = self.recover_xml_dict_key(child.tag)
				child_path = self.set_path(path, child.tag) 
				self.deserialize_child(node, child, child_path)

	def deserialize_namedtuple(
			self, 
			node, 
			child = None, 
			path = [],
			):
		self.namedtuple_to_dict(node, child, path)
		self.dict_to_namedtuple(self)

	def is_namedtuple_dict(
			self, 
			name,
			):
		if '___' in name:
			return True

	def namedtuple_name(
			self, 
			dict_name,
			):
		name = dict_name.split('___')[0]
		return name

	def namedtuple_attrs(
			self, 
			cls,
			):
		namedtuple_attr_names = []
		non_namedtuple_attr_names = []
		for attr_name in self.get_instance_dict(cls):
			if self.is_namedtuple_dict(attr_name):
				namedtuple_attr_names.append(attr_name)
			else:
				non_namedtuple_attr_names.append(attr_name)
		return namedtuple_attr_names, non_namedtuple_attr_names
			

	def namedtuple_to_dict(
			self, 
			node, 
			child = None, 
			path = [],
			):
		name = node.tag
		if self.is_namedtuple(node):
			name = self.tag_namedtuple(node) 
		if len(path) != 0:
			self.tag_child_namedtuples(child)
			null_dict = self.null_adt(
				keys=child, 
				key_attr='tag'
			)  
			self.deserialize_nested_attr(name, null_dict, path)
			for grandchild in child:
				grandchild_path = self.set_path(path, grandchild.tag) 
				self.deserialize_child(node, grandchild, grandchild_path)
		else:
			name = node.tag
			self.tag_child_namedtuples(node)
			null_dict = self.null_adt(
				keys=node, 
				key_attr='tag'
			)  
			setattr(self, name, null_dict)
			for child in node:
				child_path = self.set_path(path, child.tag) 
				self.deserialize_child(node, child, child_path)

	def dict_to_namedtuple(
			self, 
			mapping, 
			name = '',
			):
		if data_type(mapping) == 'instance':
			namedtuple_attrs, not_namedtuple_attrs \
				= self.namedtuple_attrs(mapping)  
			for key in namedtuple_attrs:
				setattr(
					mapping, 
					key.split('___')[0], 
					self.dict_to_namedtuple(
						getattr(mapping, key), 
						key
					)
				)
		if isinstance(mapping, dict):
			for key in list(mapping):
				if '___' in key:
					mapping[key.split('___')[0]] \
						= self.dict_to_namedtuple(mapping[key], key)
				else:
					mapping[key] \
						= self.dict_to_namedtuple(mapping[key], key)
			return self.namedtuple_from_mapping(mapping, name)
		return mapping

	def namedtuple_from_mapping(
			self, 
			mapping, 
			name,
			):
		if self.is_namedtuple_dict(name):
			mapping_keys = [key.split('___')[0] for key in list(mapping)]
			mapping_values = []
			for key in mapping_keys:
				mapping_values.append(mapping[key])
			mapping_ = dict(zip(mapping_keys, mapping_values))
			this_namedtuple_maker = namedtuple(name.split('___')[1], mapping_)
			return this_namedtuple_maker(**mapping_)
		else:
			return mapping
		 

	def deserialize_instance(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if len(path) != 0:
			name = node.tag 
			class_inst = self.class_instance(child)
			self.deserialize_nested_attr(name, class_inst, path)
			class_inst._deserialize(child)
		else:
			class_inst = self.class_instance(node)
			inst_attrs = InstanceAttrs(class_inst)
			inst_attrs.call_init()
			setattr(self, node.tag, class_inst)
			class_inst._deserialize(node)
		
	def attr_type(
			self, 
			node,
			):
		if len(getattr(node, 'attrib')) == 0:
			return 'NoneType'
		return node.attrib['type']

	def deserialize_child(
			self, 
			node, 
			child = None, 
			path = [],
			):
		deserialize_attr_dict = {
			'int' : self.deserialize_int,
			'float' : self.deserialize_float,
			'float64' : self.deserialize_float64,
			'float128' : self.deserialize_float128,
			'float32' : self.deserialize_float32,
			'int64' : self.deserialize_int64,
			'int32' : self.deserialize_int32,
			'complex64' : self.deserialize_complex64,
			'complex128' : self.deserialize_complex128,
			'str' : self.deserialize_str,
			'bool' : self.deserialize_bool,
			'instance' : self.deserialize_instance,
			#Added for legacy support
			'class' : self.deserialize_instance, 
			'list' : self.deserialize_list,
			'tuple' : self.deserialize_tuple,
			'ndarray' : self.deserialize_ndarray,
			'dict' : self.deserialize_dict,
			'namedtuple' : self.deserialize_namedtuple,
			'NoneType' : self.deserialize_nonetype
		}
		if len(path) != 0:
			typ = self.attr_type(child)
			return deserialize_attr_dict[typ](node, child, path)
		typ = self.attr_type(node)
		return deserialize_attr_dict[typ](node, child, path)

	def deserialize_int(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if path:
			name = node.tag
			val = int(child.text)
			return self.deserialize_nested_attr(name, val, path)
		name = node.tag
		val = int(node.text)
		setattr(self, name, val)

	def deserialize_nonetype(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if path:
			name = node.tag
			val = None 
			return self.deserialize_nested_attr(name , val, path)
		name = node.tag
		val = None 
		setattr(self, name, val)

	def deserialize_float(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if path:
			name = node.tag
			val = float(child.text)
			return self.deserialize_nested_attr(name , val, path)
		name = node.tag
		val = float(node.text)
		setattr(self, name, val)

	def deserialize_float64(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if path:
			name = node.tag
			val = np.float64(child.text)
			return self.deserialize_nested_attr(name , val, path)
		name = node.tag
		val = np.float64(node.text)
		setattr(self, name, val)

	def deserialize_float128(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if path:
			name = node.tag
			val = np.float128(child.text)
			return self.deserialize_nested_attr(name , val, path)
		name = node.tag
		val = np.float128(node.text)
		setattr(self, name, val)

	def deserialize_complex64(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if path:
			name = node.tag
			val = np.complex64(child.text)
			return self.deserialize_nested_attr(name , val, path)
		name = node.tag
		val = np.complex64(node.text)
		setattr(self, name, val)

	def deserialize_complex128(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if path:
			name = node.tag
			val = np.complex128(child.text)
			return self.deserialize_nested_attr(name , val, path)
		name = node.tag
		val = np.complex128(node.text)
		setattr(self, name, val)

	def deserialize_float32(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if path:
			name = node.tag
			val = np.float32(child.text)
			return self.deserialize_nested_attr(name , val, path)
		name = node.tag
		val = np.float32(node.text)
		setattr(self, name, val)

	def deserialize_int64(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if path:
			name = node.tag
			val = np.int64(child.text)
			return self.deserialize_nested_attr(name , val, path)
		name = node.tag
		val = np.int64(node.text)
		setattr(self, name, val)

	def deserialize_int32(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if path:
			name = node.tag
			val = np.int32(child.text)
			return self.deserialize_nested_attr(name , val, path)
		name = node.tag
		val = np.int32(node.text)
		setattr(self, name, val)

	def deserialize_bool(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if path:
			name = node.tag
			val = strcast(child.text)
			return self.deserialize_nested_attr(name, val, path)
		name = node.tag
		val = strcast(node.text)
		setattr(self, name, val)

	def set_str(
			self, 
			s,
			):
		if s == 'None':
			return ''
		return s
		
	def deserialize_str(
			self, 
			node, 
			child = None, 
			path = [],
			):
		if path:
			name = node.tag
			val = self.set_str(str(child.text))
			return self.deserialize_nested_attr(name, val, path)
		name = node.tag
		val = self.set_str(str(node.text))
		setattr(self, name, val)

	def _deserialize(
			self, 
			root, 
			path = [],
			):
		for child in root:
			if 'type' in child.attrib:
				self.deserialize_child(child)
			else:
				setattr(self, child.tag, None)
							
	def deserialize(
			self, 
			root, 
			user_input = False,
			):
		xml_root = ET.parse(root).getroot()
		if user_input:
			inst_attrs = InstanceAttrs(self)
			empty_str_attrs = []
			for attr in inst_attrs.default_instance_attrs():
				if inst_attrs.is_empty_str(attr):
					empty_str_attrs.append(attr)
			self.deserialize_user_input(
				root, 
				xml_root, 
				empty_str_attrs, 
				inst_attrs
			)
			return
		self._deserialize(xml_root)
					

'''
-----------------------------------------------------------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------------------------------------------------------------------------
'''

class Serializer(Deserialize, Serialize):
	pass




from __future__ import annotations
from collections.abc import Iterable
import re
import numpy
from typing import Union, TypeVar, NamedTuple, Sequence, Deque, Optional, List, Type, Dict, Any
from types import ModuleType

def is_instance(
        data: Any,
        ) -> bool:
    if hasattr(data, '__dict__'):
        if hasattr(data, '__module__'):
            return True

def is_namedtuple(
        data: Any,
        ) -> bool:
    typ = type(data)
    bases = typ.__bases__
    if len(bases) != 1 or bases[0] != tuple:
        return False
    fields = getattr(typ, '_fields', None)
    if not isinstance(fields, tuple):
        return False
    return all(type(ele) == str for ele in fields)

def is_dict(
        data: Any,
        ) -> bool:
    if isinstance(data, dict):
        return True 

def is_list(
        data: Any,
        ) -> bool:
    if isinstance(data, list):
        return True 

def is_ndarray(
        data: Any,
        ) -> bool:
    if isinstance(data, numpy.ndarray):
        return True 

def is_tuple(
        data: Any,
        ) -> bool:
    if isinstance(data, tuple):
        return True 

def data_type(
        data: Any,
        ) -> str: 
    if is_instance(data):
        return 'instance'
    elif is_namedtuple(data):
        return 'namedtuple'
    elif is_dict(data):
        return 'dict'
    elif is_list(data):
        return 'list'
    elif is_tuple(data):
        return 'tuple'
    elif is_ndarray(data):
        return 'ndarray'





from typing import List, Optional, Union
from userinput import UserInput
import numpy as np
import abserdes

def test_deserialize_user_input_float_list():
    user_input = UserInput()
    user_input.deserialize("xmls/user_input.xml", user_input=True)
    assert user_input.positive_float_list == [
        1.0, 2.0, 3.0
    ]
    assert user_input.positive_float_list_of_lists == [
        [1.0, 2.0, 3.0], 
        [1.0, 2.0, 3.0], 
        [1.0, 2.0, 3.0], 
    ]
    assert user_input.negative_float_list== [
        -1.0, -2.0, -3.0
    ]
    assert user_input.negative_float_list_of_lists == [
        [-1.0, -2.0, -3.0], 
        [-1.0, -2.0, -3.0], 
        [-1.0, -2.0, -3.0], 
    ]
    assert user_input.mixed_float_list == [
        1.0, -2.0, 3.0
    ]
    assert user_input.mixed_float_list_of_lists == [
        [1.0, -2.0, 3.0], 
        [1.0, -2.0, 3.0], 
        [1.0, -2.0, 3.0], 
    ]
    assert user_input.positive_int_list == [
        1, 2, 3
    ]
    assert user_input.positive_int_list_of_lists == [
        [1, 2, 3], 
        [1, 2, 3], 
        [1, 2, 3], 
    ]
    assert user_input.negative_int_list== [
        -1, -2, -3
    ]
    assert user_input.negative_int_list_of_lists == [
        [-1, -2, -3], 
        [-1, -2, -3], 
        [-1, -2, -3], 
    ]
    assert user_input.mixed_int_list == [
        1, -2, 3
    ]
    assert user_input.mixed_int_list_of_lists == [
        [1, -2, 3], 
        [1, -2, 3], 
        [1, -2, 3], 
    ]
    assert user_input.mixed_int_list_of_lists == [
        [1, -2, 3], 
        [1, -2, 3], 
        [1, -2, 3], 
    ]

    assert np.array_equal(
        user_input.positive_float64_numpy_array,
        np.array([1., 2., 3.], dtype=np.float64)
    ) == True

    assert np.array_equal(
        user_input.positive_float64_numpy_array_of_numpy_arrays,
        np.array([
            [1., 2., 3.],
            [1., 2., 3.],
            [1., 2., 3.]
        ], dtype=np.float64)
    ) == True

    assert np.array_equal(
        user_input.negative_default_float_numpy_array,
        np.array([-1., -2., -3.])
    ) == True

    assert np.array_equal(
        user_input.negative_default_float_numpy_array_of_numpy_arrays,
        np.array([
            [-1., -2., -3.],
            [-1., -2., -3.],
            [-1., -2., -3.]
        ])
    ) == True

    assert np.array_equal(
        user_input.mixed_complex128_numpy_array,
        np.array([1.+0.j, -2.+0.j, 3.+0.j], dtype=np.complex128)
    ) == True

    assert np.array_equal(
        user_input.mixed_complex128_numpy_array_of_numpy_arrays,
        np.array([
            [1.+0.j, -2.+0.j, 3.+0.j],
            [1.+0.j, -2.+0.j, 3.+0.j],
            [1.+0.j, -2.+0.j, 3.+0.j]
        ], dtype=np.complex128)
    ) == True

    assert np.array_equal(
        user_input.mixed_complex64_numpy_array,
        np.array([1.+0.j, -2.+0.j, 3.+0.j], dtype=np.complex64)
    ) == True

    assert np.array_equal(
        user_input.mixed_complex64_numpy_array_of_numpy_arrays,
        np.array([
            [1.+0.j, -2.+0.j, 3.+0.j],
            [1.+0.j, -2.+0.j, 3.+0.j],
            [1.+0.j, -2.+0.j, 3.+0.j]
        ], dtype=np.complex64)
    ) == True

    assert np.array_equal(
        user_input.positive_int32_numpy_array,
        np.array([1, 2, 3], dtype=np.int32)
    ) == True

    assert np.array_equal(
        user_input.positive_int32_numpy_array_of_numpy_arrays,
        np.array([
            [1, 2, 3],
            [1, 2, 3],
            [1, 2, 3]
        ], dtype=np.int32)
    ) == True

    assert np.array_equal(
        user_input.mixed_int64_numpy_array,
        np.array([1, -2, 3], dtype=np.int64)
    ) == True

    assert np.array_equal(
        user_input.mixed_int64_numpy_array_of_numpy_arrays,
        np.array([
            [1, -2, 3],
            [1, -2, 3],
            [1, -2, 3]
        ], dtype=np.int64)
    ) == True

    assert np.array_equal(
        user_input.negative_int_float_cast_numpy_array,
        np.array([-1., -2., -3.])
    ) == True

    assert np.array_equal(
        user_input.negative_int_float_cast_numpy_array_of_numpy_arrays,
        np.array([
            [-1., -2., -3.],
            [-1., -2., -3.],
            [-1., -2., -3.]
        ])
    ) == True

    assert user_input.str_list == [
        "Langevin",
        "OpenMM"
    ]


    assert user_input.str_expr_list == [
        "step(distance(g1, g2)^2 - radius^2",
        "0.5*k*(theta-theta0)^2",
    ]
    
    



import numpy as np
from collections import deque
import re
import sys
from typing import List, Tuple, Optional, Union, Dict, Type


class ListStrCastNode(object): 

	def __init__(
			self,
			start,
			end,
			parent = None,
			path = [],
			):
		self.start = start
		self.end = end
		self.parent = parent
		self.utils = StrCastUtils()
		self.path = path 

	def top_child(
			self,
			children,
			):
		return list(children)[-1] 

	def pop_child(
			self,
			children,
			pos,
			):
		return children.pop(pos) 

	def is_child(
			self,
			start1,
			end1,
			start2,
			end2,
			):
		if start1 > start2 and end1 < end2:
			return True
		return

	def is_leaf(
			self,
			children,
			pos,
			):
		if self.top_child(children) == pos:
			return True
		return

	def is_chars_leaf(
			self,
			chars_dict,
			pos,
			):
		if self.top_child(chars_dict) == pos:
			return True
		return

	def get_children(
			self,
			internal_nodes = True,
			leaves = True,
			):
		if internal_nodes and leaves:
			return [
				child for name, child in self.__dict__.items() 
					if 'child' in name or 'leaf' in name
			]
		elif internal_nodes and not leaves:
			return [
				child for name, child in self.__dict__.items() 
					if 'child' in name
			]
		elif leaves and not internal_nodes:
			return [
				child for name, child in self.__dict__.items() 
					if 'leaf' in name
			]
		return

	def child_names(
			self, 
			node,
			):
		return [
			name for name in self.__dict__ 
				if 'child' in name
		]

	def leaf_names(
			self,
			node,
			):
		return [
			name for name in self.__dict__
				if 'leaf' in name
		]

	def child_path(self):
		path = [] 
		if self.path:
			path = [pos for pos in self.path] 
		path.append(len(self.get_children(self)))
		return path

	def add_child(
			self, 
			start,
			end,
			):
		if self.child_names(self):
			name = (
				'child'
				+ str(int(re.findall(
					r'\d+', 
					self.child_names(self)[-1])[0]) 
				+ 1)
			)
		else:
			name = 'child0'
		setattr(
			self, name, ListStrCastNode(
				start, end, self,
				self.child_path()
			)
		) 
		return getattr(self, name) 

	def add_leaf(
			self,
			start,
			end, 
			s,
			chars = False
			):
		name = None
		if self.leaf_names(self):
			name = (
				'leaf'
				+ str(int(re.findall(
					r'\d+', 
					self.leaf_names(self)[-1])[0])
				+ 1)
			)
		else:
			name = 'leaf0'
		leaf = self.utils.cleanup_str(
			s, start + 1,
			end, 'list'
		)
		if chars:
			setattr(self, name, tuple(leaf))
		else:
			setattr(self, name, leaf)
		return getattr(self, name) 

	def is_numpy_dtype(
			self,
			val: str,
			) -> bool:
		numpy_dtype_suffixes = (
			'_INT32',
			'_INT64',
			'_FLOAT32',
			'_FLOAT64',
			'_COMPLEX64',
			'_COMPLEX128',
		)
		for numpy_dtype_suffix in numpy_dtype_suffixes:
			if numpy_dtype_suffix in val:
				return True
		return False

class StrCastUtils(object):

	def is_numpy_dtype(
			self,
			val: str,
			) -> bool:
		numpy_dtype_suffixes = (
			'_INT32',
			'_INT64',
			'_FLOAT32',
			'_FLOAT64',
			'_FLOAT128',
			'_COMPLEX64',
			'_COMPLEX128',
		)
		for numpy_dtype_suffix in numpy_dtype_suffixes:
			if numpy_dtype_suffix in val:
				return True
		return False

	def parse_numpy_dtype(
			self,
			val: str,
			) -> Tuple[str]:
		dtype_name = \
			val.partition('_')[-1].replace('_', '').lower() 
		parsed_val = val.partition('_')[0]
		return parsed_val, dtype_name

	def xml_str_cast(
			self,
			val,
			):
		if type(val.text) == type(None):
			return	 
		else:
			if val.text.lower() == 'true':
				return True 
			elif val.text.lower() == 'false':
				return False 
			elif '[' in val.text:
				cast = StrCast()
				return cast.str_to_list(val)
			elif self.is_numpy_dtype(val.text):
				parsed_str_val, dtype_name = \
					self.parse_numpy_dtype(val.text)
				return self.__str_cast(val.text, dtype_name)			 
			elif val.text == '.' or re.search('[a-zA-Z]', val.text):
				return self.__str_cast(val.text, 'str')
			elif re.search('[\.]', val.text):
				return self.__str_cast(val.text, 'float')
			else:
				return self.__str_cast(val.text, 'int')

	def std_str_cast(
			self,
			val,
			):
		if type(val) == type(None):
			return	 
		else:
			if val.lower() == 'true':
				return True 
			elif val.lower() == 'false':
				return False 
			elif '[' in val:
				cast = StrCast()
				return cast.str_to_list(val)
			elif self.is_numpy_dtype(val):
				parsed_val, dtype_name = \
					self.parse_numpy_dtype(val)
				return self.__str_cast(parsed_val, dtype_name)			   
			elif val == '.' or re.search('[a-zA-Z]', val):
				return self.__str_cast(val, 'str')
			elif re.search('[\.]', val):
				return self.__str_cast(val, 'float')
			else:
				return self.__str_cast(val, 'int')

	def str_cast(
			self, 
			val,
			):
		if hasattr(val, 'text'):
			return self.xml_str_cast(val)
		else:
			return self.std_str_cast(val)

	def __str_cast(
			self, 
			str_val,
			type_str,
			):
		type_lookup = {
			'int' : int,
			'float' : float,
			'str' : str,
			'int32' : np.int32,
			'int64' : np.int64,
			'float32' : np.float32,
			'float64' : np.float64,
			'float128' : np.float128,
			'complex64' : np.complex64,
			'complex128' : np.complex128
		} 
		return type_lookup[type_str](str_val)

	def except_the_kitchen_sink_regex(
			self, 
			str_struct
			):
		struct = re.findall(
			r'\"[^\"]*\"|\S*[A-Za-z]\S*[0-9]\S*|'
			'[0-9]?\S*[A-Za-z]\S*|'
			'-?\.?\d+\.?\d*|\d+',
			str_struct
		)
		return struct

	def str_cast_list_elements(
			self,
			str_list,
			):
		parsed_str_list = self.except_the_kitchen_sink_regex(
			str_list
		)
		return self.cleanup_str_list(
			parsed_str_list 
		)

	def cleanup_str_list(
			self,
			parsed_str_list,
			) -> List[str]:
		parsed_str_list = [
			list_element if '"' in list_element 
			else list_element.replace(',', '') 
				for list_element in parsed_str_list
		]
		parsed_str_list = [
			list_element.replace('"', '') for list_element
			in parsed_str_list
		]
		parsed_str_list = [
			self.str_cast(list_element) for list_element
			in parsed_str_list
		]
		return parsed_str_list
		

	def str_cast_dict_elements(
			self,
			str_dict,
			): #Come back to this later
		return	

#get rid of *args
	def str_cast_struct_elements(
			self,
			str_struct,
			*args,
			):
		str_cast_struct_dict = {
			'list' : self.str_cast_list_elements(str_struct), 
			'dict' : self.str_cast_dict_elements(str_struct),
		}
		struct_type = args[0]
		return str_cast_struct_dict[struct_type]

#get rid of *args
	"""
	def find_matching(
			self,
			s,
			*args,
			):
		matching_dict = {}
		stack = []
		start, end = args
		for pos, i in enumerate(s):
			if i == start:
				stack.append(pos)
			elif i == end:
				matching_dict[stack.pop()] = pos
		return matching_dict
	"""

	def is_start_end(self, char, start, end, matching_dict):
		if start == end:
			if char == start:
				if None in matching_dict.values():
					return 'end'
				return 'start'
			return ''
		if char == start:
			return 'start'
		elif char == end:
			return 'end'
		return ''

	def find_matching(
			self, 
			s,
			start,
			end,
			):
		matching_dict = {}
		stack = []
		for pos, i in enumerate(s):
			if self.is_start_end(i, start, end, matching_dict) == "start":
				matching_dict[pos] = None
				stack.append(pos)
			elif self.is_start_end(i, start, end, matching_dict) == "end":
				matching_dict[stack.pop()] = pos
		return matching_dict

	def num_lists(
			self,
			s,
			):
		return self.num_structs(s, 'list')

	def num_dicts(
			self,
			s,
			):
		return self.num_structs(s, 'dict')

	def num_structs(
			self,
			s,
			struct_type,
			):
		num_structs_dict = { 
			'list' : self.find_matching(s, '[', ']'), 
			'dict' : self.find_matching(s, '{', '}')
		}
		return len(num_structs_dict[struct_type])

	def nested_list_insert(
			self,
			lst,
			val,
			path,
			):
		if isinstance(path, list):
			path = deque(path)
		self.__nested_list_insert(lst, val, path)
		return lst

	def __nested_list_insert(
			self,
			lst,
			val,
			path,
			):
		if path:
			pos = path.popleft()
			return self.__nested_list_insert(lst[pos], val, path)
		lst.append(val)

	def sort_dict(
			self,
			d,
			*args,
			**kwargs,
			):
		args = list(args)
		for index, arg in enumerate(args):
			if arg == 'key':
				args[index] = 0
			if arg == 'value':
				args[index] = 1
		kwargs_keys = [key for key, value in kwargs.items()]
		return {
			key : value for key, value in sorted(
				d.items(), key = lambda item : item[args[0]], 
				reverse = kwargs[kwargs_keys[0]]
			)
		}

	def cleanup_str(
			self, 
			s, 
			start, 
			end, 
			*args,
			):
		return self.str_cast_struct_elements(s[start : end], *args)

	def chars(
			self, 
			s, 
			*args,
			):
		chars_dict = {}
		prev_bracket = {}
		for index, char in enumerate(s):
			if char == args[0]:
				if prev_bracket:
					start, end = list(prev_bracket)[-1], index
					if prev_bracket[start] == args[0]:
						if (end - start) != 1:
							chars_dict[start] = self.cleanup_str(
								s, start, end,
								args[2]
							)
					elif prev_bracket[start] == args[1]:
						if (end - start) != 3:
							chars_dict[start] = self.cleanup_str(
									s, start, end,
									args[2]
								)
				prev_bracket = {index : args[0]} 
			elif char == args[1]:
				start, end = list(prev_bracket)[-1], index
				if prev_bracket[start] == args[1]:
					if (end - start) != 1:
						chars_dict[start] = self.cleanup_str(
							s, start, end,
							args[2]
						)
				prev_bracket = {index : args[1]} 
		return chars_dict

	
class StrCast(object):

	def __init__(self):
		self.utils = StrCastUtils()

	def adjust_recursion_limit(
			self,
			s,
			):
		if self.utils.num_lists(s) > 100:
			sys.setrecursionlimit(int(10e8))

	def start_end(
			self, 
			q_s, 
			pos = None,
			):
		if not pos:
			start = list(q_s[-1])[0]
			end = q_s[-1][start]
			return start, end 
		start = list(q_s[pos])[0]
		end = q_s[pos][start]
		return start, end 
		
	def list_init_queue(
			self, 
			s,
			):
		return deque([
			{start : end} for start, end in self.utils.sort_dict(
				self.utils.find_matching(s, '[', ']'),
				'key', reverse = True).items()
		])

	def list_init_children(
			self, 
			s,
			):
		return self.utils.sort_dict(
			self.utils.find_matching(s, '[', ']'),
			'value', reverse = True
		)

	def list_init_stack(
			self,
			s,
			init_queue = None,
			):
		if init_queue:
			return deque([init_queue.pop()])
		else:
			return deque()

	def list_str_tree(
			self,
			s,
			**qscpn,
			):
		if not qscpn:
			self.adjust_recursion_limit(s)
			qscpn['queue'] = self.list_init_queue(s)
			if len(qscpn['queue']) == 1:
				qs, qe = self.start_end(qscpn['queue']) 
				node = ListStrCastNode(qs, qe)
				node.add_leaf(qs, qe, s)
				return node
			qscpn['stack'] = self.list_init_stack(s, qscpn['queue'])
			qscpn['children'] = self.list_init_children(s) 
			qscpn['pos'] =	self.start_end(qscpn['queue'])[0]
			qscpn['node'] = ListStrCastNode(*self.start_end(qscpn['stack']))
			qscpn['node'].pop_child(qscpn['children'], 0)
		queue, stack, children, pos, node = qscpn.values()
		qs, qe = self.start_end(queue) 
		ss, se = self.start_end(stack) 
		if len(queue) > 1:
			if se > qe:
				if node.is_leaf(children, pos):
					node.pop_child(children, qs)
					node.add_leaf(qs, qe, s)
					queue.pop()
					pos = self.start_end(queue)[0]
					qscpn = dict(zip(
						qscpn.keys(), (
							queue,
							stack, 
							children, 
							pos, 
							node,
						)
					))
					self.list_str_tree(s, **qscpn)
				else:
					node.pop_child(children, qs)
					child = node.add_child(qs, qe)
					stack.append(queue.pop())
					pos = self.start_end(queue)[0]
					qscpn = dict(zip(
						qscpn.keys(), (
							queue, 
							stack, 
							children,
							pos,
							child,
						)
					))
					self.list_str_tree(s, **qscpn)
			elif qe > se:
				stack.pop()
				qscpn = dict(zip(
					qscpn.keys(), (
						queue, 
						stack, 
						children, 
						pos, 
						node.parent,
					)
				))
				self.list_str_tree(s, **qscpn)
		if children:
			if node.is_child(
					qs, qe, ss, 
					se
				):
				if node.is_leaf(children, pos):
					node.pop_child(children, qs)
					node.add_leaf(qs, qe, s)
					queue.pop()
			else:
				stack.pop()
				qscpn = dict(zip(
					qscpn.keys(), ( 
						queue, 
						stack,
						children, 
						pos, 
						node.parent
					)
				))
				self.list_str_tree(s, **qscpn)
		return node

	def list_str_tree_to_list(
			self, 
			tree, 
			lst,
			):
		self.__list_str_tree_to_list(tree, lst)
		return lst

	
	def __list_str_tree_to_list(
			self, 
			node, 
			lst,
			):
		if node.get_children():
			for child in node.get_children():
				if isinstance(child, type(node)):
					self.utils.nested_list_insert(lst, [], node.path)
					self.__list_str_tree_to_list(child, lst)
				if isinstance(child, list):
					self.utils.nested_list_insert(lst, child, node.path)
		
	def str_to_list(
			self, 
			s,
			):
		tree = self.list_str_tree(s)
		lst = []
		if self.utils.num_lists(s) == 1:
			return self.list_str_tree_to_list(tree, lst)[0]
		return self.list_str_tree_to_list(tree, lst)
	
def strcast(
		s,
		):
	str_cast = StrCast()
	return str_cast.utils.str_cast(s)




"""
abserdes
A Basic Serializer and Deserializer that makes XML serialization and
deserialization in Python absurdly easy.
"""

import sys
import os
from setuptools import setup, find_packages, Extension


setup(
    name='abserdes',
    author='Andy Stokely',
    author_email='amstokely@ucsd.edu',
    license='MIT',
    description="A Basic Serializer and Deserializer that makes XML"
        + "serialization and deserialization in Python absurdly easy.",
    keywords='xml, serialize, deserialize, serializer',
    url='https://github.com/astokely/abserdes',
    packages=find_packages(),
    install_requires=[
		"numpy", "pytest", 
		"data-science-types", "cython"
	],              
    platforms=['Linux',
                'Unix',],
    python_requires=">=3.8",          
)



from __future__ import annotations
from typing import Union, TypeVar, NamedTuple, Sequence, Deque, Optional, List, Type, Dict, Any
from types import ModuleType

from .strcast import strcast



CLASS = TypeVar('CLASS')

#Generic class instance type alias
INSTANCE = TypeVar('INSTANCE')

class UserInputNode(object):
	
	def __init__(self,
			xml_node: ET.Element, 
			name: str, 
			type_: str, 
			parent: Optional[UserInputNode] = None,
			) -> None:
		self.xml_node = xml_node
		self.name = name
		self.type_ = type_ 
		self.parent = parent
		self.val = None

	def data_type(
			self, 
			xml_node,
			):
		if not list(xml_node):
			return 'non_instance'
		else:
			class_children = []
			for child in xml_node:
				if getattr(child, 'attrib'):
					class_children.append(child)
			if len(class_children) == len(xml_node):
				return 'instance_list'
			else:
				return 'instance'

	def get_name(
			self, 
			xml_node,
			):
		return xml_node.tag 

	def add_child(
			self, 
			xml_node,
			):
		name = self.get_name(xml_node)
		type_ = self.data_type(xml_node)
		child = UserInputNode(
			xml_node, 
			name, type_, 
			self
		)
		if type_ == 'pdt':
			child.data = strcast(xml_node.text)
		setattr(self, name, child)
		return getattr(self, name)



from __future__ import annotations
from typing import Union, TypeVar, NamedTuple, Sequence, Deque, Optional, List, Type, Dict, Any
from types import ModuleType
import copy
import inspect
import ast
import textwrap
import importlib
from .serializerutils import SerializerUtils
from functools import lru_cache as lru


#Generic class type alias
CLASS = TypeVar('CLASS')

#Generic class inst type alias
INSTANCE = TypeVar('INSTANCE')

class InstanceAttrs(SerializerUtils):

    def __init__(
            self,
            inst: INSTANCE,
            ) -> None:
        self.inst = inst 

    def init_src_code(self) -> str:
        return textwrap.dedent(
            inspect.getsource(self.inst.__init__))

    def instance_attrs_ast(
            self,
            init_src_code: str,
            ) -> List[ast.Assign]:
        inst_attrs_ast = ast.parse(init_src_code).body[0].body
        return [
            node for node in inst_attrs_ast 
            if isinstance(node, ast.Assign)
        ]


    def default_instance_attr_names(
            self,
            inst_attrs_ast: List[ast.Assign],
            ) -> List[str]:
        return [
            node.targets[0].attr for node in inst_attrs_ast
            if isinstance(node, ast.Assign)
        ]

    def default_instance_attr_values(
            self,
            inst_attrs_ast: List[ast.Assign],
            ):
        for node in inst_attrs_ast:
            value = self.get_instance_dict(node.value, values=True)[0]
            yield value

    def default_instance_attrs(self) -> Dict[str, Any]:
        init_src_code = self.init_src_code() 
        inst_attrs_ast = self.instance_attrs_ast(init_src_code) 
        default_inst_attr_names = \
            self.default_instance_attr_names(inst_attrs_ast)
        default_inst_attr_values = \
            list(self.default_instance_attr_values(inst_attrs_ast))
        return dict(zip(
            default_inst_attr_names, default_inst_attr_values
            ))

    def is_empty_str(
            self,
            default_inst_attr: str,
            ) -> Union[bool, None]:
        default_inst_attrs = self.default_instance_attrs()
        if default_inst_attrs[default_inst_attr] == '':
            return True

    def init_args(self):
        class_ = type(self.inst)
        signature = inspect.signature(self.inst.__init__)
        return [
            arg_name for arg_name in list(signature.parameters)
        ]

    def instance_attrs(self):
        return copy.deepcopy(self.get_instance_dict(self.inst))

    def map_instance_attrs_to_init_args(self):
        inst_attr_names = list(self.instance_attrs())
        args = []
    
        for arg in self.init_args():
            if arg in inst_attr_names:
                args.append(getattr(self.inst, arg))
            else:
                return
        return tuple(args)

    def call_init__(
            self,
            args, 
            ):
        inst_attrs = self.instance_attrs()
        self.inst.__init__(*args)
        for attr_name, attr_value in inst_attrs.items():
            setattr(self.inst, attr_name, attr_value)  
        return

    def call_init(self):
        args = self.map_instance_attrs_to_init_args()
        if isinstance(args, tuple):
            self.call_init__(args)




from typing import List, Optional, Union
import pytest
from userinput import UserInput
import numpy as np
import abserdes
from conftest import Dictionary

@pytest.mark.dictionary
def test_dictionary_serialization(
			tmp_test_files_dir
):
	dictionary = Dictionary()
	dictionary.dictionary = {
		'a' : 1, 
		'b' : 2, 
		'c' : 3
	}
	dictionary.serialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	deserialized_dictionary = Dictionary()
	deserialized_dictionary.deserialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	assert ( 
		deserialized_dictionary.dictionary 
		== dictionary.dictionary
	)

@pytest.mark.dictionary
def test_tuple_key_dictionary_serialization(
        tmp_test_files_dir
):
	dictionary = Dictionary()
	dictionary.tuple_key_dictionary = {
		(1, 2, 3) : 1, 
		(4, 5, 6) : 2, 
		(7, 8, 9) : 3
	}
	dictionary.serialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	deserialized_dictionary = Dictionary()
	deserialized_dictionary.deserialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	assert ( 
		deserialized_dictionary.tuple_key_dictionary 
		== dictionary.tuple_key_dictionary
	)

@pytest.mark.dictionary
def test_int_key_dictionary_serialization(
        tmp_test_files_dir
):
	dictionary = Dictionary()
	dictionary.int_key_dictionary = {
		1 : 1, 
		4 : 2, 
		7 : 3
	}
	dictionary.serialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	deserialized_dictionary = Dictionary()
	deserialized_dictionary.deserialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	assert ( 
		deserialized_dictionary.int_key_dictionary 
		== dictionary.int_key_dictionary
	)

@pytest.mark.dictionary
def test_float128_key_dictionary_serialization(
        tmp_test_files_dir
):
	dictionary = Dictionary()
	dictionary.float128_key_dictionary = {
		np.float128(1.23) : 1, 
		np.float128(4.56) : 2, 
		np.float128(7.89) : 3
	}
	dictionary.serialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	deserialized_dictionary = Dictionary()
	deserialized_dictionary.deserialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	assert (
		deserialized_dictionary.float128_key_dictionary 
		== dictionary.float128_key_dictionary
	)

@pytest.mark.dictionary
def test_nested_dictionary_serialization(
        tmp_test_files_dir
):
	dictionary = Dictionary()
	dictionary.nested_dictionary = {
		'a' : {'a' : 1}, 
		'b' : {'b' : 2}, 
		'c' : {'c' : 3}
	}
	dictionary.serialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	deserialized_dictionary = Dictionary()
	deserialized_dictionary.deserialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	assert (
		deserialized_dictionary.nested_dictionary 
		== dictionary.nested_dictionary
	)

@pytest.mark.dictionary
def test_float_key_dictionary_serialization(
        tmp_test_files_dir
):
	dictionary = Dictionary()
	dictionary.float_key_dictionary = {
		1.23 : 1, 
		4.56 : 2, 
		7.89 : 3
	}
	dictionary.serialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	deserialized_dictionary = Dictionary()
	deserialized_dictionary.deserialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	assert (
		deserialized_dictionary.float_key_dictionary 
		== dictionary.float_key_dictionary
	)

@pytest.mark.dictionary
def test_int_key_dictionary_serialization(
        tmp_test_files_dir
):
	dictionary = Dictionary()
	dictionary.int_key_dictionary = {
		1 : 1, 
		4 : 2, 
		7 : 3
	}
	dictionary.serialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	deserialized_dictionary = Dictionary()
	deserialized_dictionary.deserialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	assert (
		deserialized_dictionary.int_key_dictionary 
		== dictionary.int_key_dictionary
	)

@pytest.mark.dictionary
def test_float64_key_dictionary_serialization(
        tmp_test_files_dir
):
	dictionary = Dictionary()
	dictionary.float64_key_dictionary = {
		np.float64(1.23) : 1, 
		np.float64(4.56) : 2, 
		np.float64(7.89) : 3
	}
	dictionary.serialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	deserialized_dictionary = Dictionary()
	deserialized_dictionary.deserialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	assert (
		deserialized_dictionary.float64_key_dictionary 
		== dictionary.float64_key_dictionary
	)

@pytest.mark.dictionary
def test_nested_tuple_key_dictionary_serialization(
        tmp_test_files_dir
):
	dictionary = Dictionary()
	dictionary.nested_tuple_key_dictionary = {
		(1, 2, 3) : {6.6658 : {(9.332, np.float128(5.223)) : 'abc'}}, 
		(4, np.float64(5.2), 6) : {
			np.float64(10.6658) : {
				(9.332, np.float128(5.223)) : 'a81981bc'
			}
		}, 
		(4, 5, 6) : {
			np.float64(10.6658) : {
				(9.332, np.float128(5.223)) : 'a81981bc'
			}
		}, 
	}
	dictionary.serialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	deserialized_dictionary = Dictionary()
	deserialized_dictionary.deserialize(str(
		tmp_test_files_dir /
		"dictionary.xml"
	))
	assert (
		deserialized_dictionary.nested_tuple_key_dictionary == 
		dictionary.nested_tuple_key_dictionary
	)












